/*----------------------------------------------------------------------
  usbclock2 by Jack Botner

  This program provides host support for a USB attached ATmega8 LCD
  display which may be used to display date and time.
  USB support is provided by a FTDI UB232R USB-serial module connected
  to the ATmega8. In this version, the D2XX interface is used.

  Command line parameter required: COMx

  jb20110225 - A problem that occurred occasionally under Win7 32bit
               and more often under Win7 64bit, at startup the program 
			   can't find the FTDI driver and quits. Apparently the 
			   program is being run from the startup menu before the 
			   drivers are initialized, so the solution appears to be 
			   wait and retry a few times.
  jb20131218 - Added commands 'U' (UTC time) and 'L' (local time) from
               remote.
  jb20150608 - Added startup synchronization routine.
  jb20150822 - Reorganized checksum calculation, so I could turn off
               the high-order bit. This is to solve a communication
			   failure after updating the FTDI driver. I guess the
			   new driver doesn't believe '8 bits no parity' anymore.
----------------------------------------------------------------------*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "usbclock2.h"
#include "ftd2xx.h"
#include "resource.h"

#define NR_TRIES       10		// times to try initialization

// Global variables

static HFONT     hFont=NULL;
static BOOL      bIconAdded=FALSE;
static HICON     hIconApp;
static HWND      hwndDlg=NULL;
static int       iVcomPort=1;
static int       iDateTimeFormat=DATETIMELOCAL;
static FT_HANDLE ftHandle=NULL;

static char *pcMonth[13] =
{
  "Err", "Jan.", "Feb.", "Mar.", "Apr.", "May",  "June",
  "July", "Aug.", "Sep.", "Oct.", "Nov.", "Dec."
};

static char *pcDay[8] =
{
  "SU", "MO", "TU", "WE", "TH", "FR", "SA", "??"
};

// Internal functions

static unsigned char checksum( unsigned char *pucBuffer, int nBytes );
static BOOL dialogInit( HWND hwnd, LPARAM lParam );
static int getComPortHandle( BOOL bReport );
static LRESULT APIENTRY MainDlgMsgs( HWND hwnd, ULONG msg,
                                     WPARAM wParam, LPARAM lParam );
static BOOL MyTaskBarIcon( HWND hwnd, HICON hIcon );
static BOOL MyTaskBarDeleteIcon( HWND hwnd );
static void On_MYWM_NOTIFYICON( HWND hwnd, WPARAM wParam, LPARAM lParam );
static BOOL processTimer( HWND hwnd );
static BOOL synchronize_avr( FT_HANDLE ftHandle );

/*----------------------------------------------------------------------
  WinMain(...)
----------------------------------------------------------------------*/
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrev,
                    LPSTR lpArgs, int nCmdShow )
{
  int           iRc, nn;
  BOOL          bRc, bReport;
  DWORD         dwRc;
  char          cMsg[92];
  LOGFONT       logFont;
  HANDLE        hMutexPgm;
  FT_STATUS     ftStat;
  DWORD         dwArg1, dwArg2, dwBytesWritten;
  unsigned char ucBuf[16];

  // To make sure only one instance of this program runs
  hMutexPgm = CreateMutex( NULL, FALSE, APPNAME );
  dwRc = GetLastError();
  if ( !hMutexPgm )
  {
    sprintf( cMsg, "CreateMutex (PGM) failed, rc = %u", dwRc );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    return 0;
  }

  if ( dwRc == ERROR_ALREADY_EXISTS )
  {
    sprintf( cMsg, "%s is already running." , APPNAME "PGM" );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    return 0;
  }

  // Get the com port specification from the command line
  nn = strlen( lpArgs );
  if ( nn == 4 && !_memicmp( lpArgs, "com", 3 ) )
  {
    nn = atoi( lpArgs+3 );
    if ( nn > 0 && nn < 12 )
      iVcomPort = nn;
  }

  // Get the number of USB devices available
  ftStat = FT_ListDevices( &dwArg1, &dwArg2, FT_LIST_NUMBER_ONLY );
  if ( ftStat != FT_OK )
  {
    sprintf( cMsg, "FT_ListDevices failed, rc = %d", (unsigned) ftStat );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    return 0;
  }
  if ( !dwArg1 )
  {
    MessageBox( NULL, "FT_ListDevices reports no devices found.",
                APPNAME, MB_OK | MB_ICONERROR );
    bRc = CloseHandle( hMutexPgm );
    return 0;
  }

  // begin jb20110225

  for ( nn=0; nn<NR_TRIES; ++nn )
  {
	bReport = ( nn < ( NR_TRIES-1 ) ) ? FALSE : TRUE; 
    iRc = getComPortHandle( bReport );
    if ( iRc == 1 ) break;                      // successful
	if ( iRc == -1 )
	{
	  bRc = CloseHandle( hMutexPgm );
	  return 0;									// fail no recovery
	}

	Sleep( 1000 );								// wait and try again
  }

  if ( nn == NR_TRIES )
  {
	  bRc = CloseHandle( hMutexPgm );
	  return 0;									// fail too many times
  }
  // end jb20110225

  // Ok guess we got it....
  ftStat = FT_SetBaudRate( ftHandle, 9600 );    // what our ATmega8 uses
  if ( ftStat == FT_OK )
    ftStat = FT_SetDataCharacteristics( ftHandle, FT_BITS_8, FT_STOP_BITS_1,
                                        FT_PARITY_NONE );
  if ( ftStat == FT_OK )
    ftStat = FT_SetFlowControl( ftHandle, FT_FLOW_NONE, 0, 0 );
  if ( ftStat != FT_OK )
  {
    sprintf( cMsg, "FT_setup failed, rc = %u", (unsigned) ftStat );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    ftStat = FT_Close( ftHandle );
    bRc = CloseHandle( hMutexPgm );
    return 0;
  }

  // Synchronize with the remote application jb20150608
  while ( 1 )
  {
    bRc = synchronize_avr( ftHandle );
    if ( bRc )
	  break;

    iRc = MessageBox( NULL, "Failed to synchronize with clock. Retry?", APPNAME, 
                      MB_YESNO | MB_ICONERROR );  
    if ( iRc == IDYES )
      continue;
    ftStat = FT_Close( ftHandle );
    bRc = CloseHandle( hMutexPgm );
    return 0;
  }

  // Load application icon
  hIconApp = LoadIcon( hInstance, MAKEINTRESOURCE( IDI_ICON ) );

  // Create a fixed pitch font for the text window
  memset( &logFont, '\0', sizeof( LOGFONT ) );
  logFont.lfHeight = 14;
  logFont.lfWidth  = 10;
  logFont.lfWeight = FW_NORMAL;
  logFont.lfCharSet = ANSI_CHARSET;
  logFont.lfOutPrecision = OUT_DEFAULT_PRECIS;
  logFont.lfClipPrecision = CLIP_DEFAULT_PRECIS;
  logFont.lfQuality = DEFAULT_QUALITY;
  logFont.lfPitchAndFamily = FIXED_PITCH | FF_DONTCARE;
  strcpy( logFont.lfFaceName, "Courier New" );
  hFont = CreateFontIndirect( &logFont );
  if ( !hFont )
  {
    dwRc = GetLastError();
    sprintf( cMsg, "CreateFontIndirect failed, rc = %u", dwRc );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONWARNING );
  }

  iRc = DialogBoxParam( hInstance, "USBCLOCK2", HWND_DESKTOP,
                       (DLGPROC) MainDlgMsgs, (LPARAM) nCmdShow );
  dwRc = GetLastError();

  if ( hFont )
    bRc = DeleteObject( hFont );

  if ( ftHandle )
  {
    ucBuf[0] = 'Z';               // shutdown
    ucBuf[1] = 3;
    ucBuf[2] = ucBuf[0] + ucBuf[1];
    ftStat = FT_Write( ftHandle, ucBuf, 3, &dwBytesWritten );

    ftStat = FT_Close( ftHandle );
  }

  bRc = CloseHandle( hMutexPgm );

  return 0;
}

/*----------------------------------------------------------------------
   MainDlgMsgs(...)
----------------------------------------------------------------------*/
static LRESULT APIENTRY MainDlgMsgs( HWND hwnd, ULONG msg,
                                     WPARAM wParam, LPARAM lParam )
{
  BOOL          bRc, bRcode=FALSE;
  DWORD         dwBytesWritten;
  FT_STATUS     ftStat;

  switch ( msg )
  {
    case WM_INITDIALOG:
      bRc = dialogInit( hwnd, lParam );
      break;

    case WM_CLOSE:
      EndDialog( hwnd, 0 );
      break;

    case WM_DESTROY:
      bRc = KillTimer( hwnd, CLOCK_TIMER_ID );

      if ( bIconAdded )
        bRc = MyTaskBarDeleteIcon( hwnd );
      break;

    case WM_COMMAND:
      switch ( LOWORD( wParam ) )
      {
        case IDC_TEST:
          {
            unsigned char ucBuf[24];
            char          cMsg[256];

            // To sound the beeper for testing
            // Build the beeper message for the micro
            ucBuf[0] = 'M';               // get id cmd
            ucBuf[1] = 3;
            ucBuf[2] = checksum( ucBuf, 2 );

            ftStat = FT_Write( ftHandle, ucBuf, 3, &dwBytesWritten );
            if ( ftStat != FT_OK )
            {
              sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
              listMessage( cMsg, 0 );
            }
          }
          break;

        case IDCANCEL:
          EndDialog( hwnd, 0 );
          break;
      }
      break;

    case WM_TIMER:
      if ( wParam == CLOCK_TIMER_ID )
        processTimer( hwnd );
      break;

    case WM_SIZE:
      if ( wParam == SIZE_MINIMIZED )
      {
        bRc = ShowWindow( hwnd, SW_HIDE /* SW_MINIMIZE */ );
        bRc = MyTaskBarIcon( hwnd, hIconApp );
        bRcode = FALSE;  // message processed
        break;
      }

      bRcode = TRUE;   // message not processed
      break;

    case WM_QUERYENDSESSION:
      // We are being asked if its OK to end the session
      bRcode = TRUE;            // reply yes
      break;

    case WM_ENDSESSION:
      // The session is ending.
      EndDialog( hwnd, 0 );
      bRcode = FALSE;           // message processed
      break;

    case UM_NOTIFYICON:
      // This is my message for notifications from the taskbar icon.
      On_MYWM_NOTIFYICON( hwnd, wParam, lParam);
      break;

  }   // end switch

  return (LRESULT) bRcode;
}

/*----------------------------------------------------------------------
  dialogInit(...)

  returns: TRUE = OK, FALSE = Error
----------------------------------------------------------------------*/
static BOOL dialogInit( HWND hwnd, LPARAM lParam )
{
  int         cx, cy, xx, yy;
  LRESULT     lRc;
  BOOL        bRc;
  RECT        rectDlg, rectScrn;
  UINT        uiRc;
  DWORD       dwRc;
  DWORD       dwBytesWritten;
  FT_STATUS   ftStat;
  char        cMsg[256];
  unsigned char ucBuf[32];

  hwndDlg = hwnd;

  if ( hFont )
    lRc = SendDlgItemMessage( hwnd, IDC_MSGS, WM_SETFONT,
                              (WPARAM) hFont, (LPARAM) FALSE );

  // Center the window
  GetWindowRect( GetDesktopWindow(), &rectScrn );
  GetWindowRect( hwnd, &rectDlg );  // returns rectangle relative to screen
  cx = rectDlg.right - rectDlg.left;
  cy = rectDlg.bottom - rectDlg.top;
  xx = ( rectScrn.right - cx ) / 2;
  yy = ( rectScrn.bottom - cy ) / 2;
  bRc = SetWindowPos( hwnd, HWND_TOP, xx, yy, cx, cy, (UINT) lParam /* SWP_SHOWWINDOW */ );

  uiRc = SetTimer( hwnd, CLOCK_TIMER_ID, 1000, NULL );
  if ( !uiRc )
  {
    dwRc = GetLastError();
    sprintf( cMsg, "SetTimer failed, windows rc = %u", dwRc );
    listMessage( cMsg, 0 );
    return FALSE;
  }

  // Build the get-ID message for the micro
  ucBuf[0] = 'M';               // get ID
  ucBuf[1] = 3;
  ucBuf[2] = checksum( ucBuf, 2 );

  ftStat = FT_Write( ftHandle, ucBuf, 3, &dwBytesWritten );
  if ( ftStat != FT_OK )
  {
    bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
    sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
    listMessage( cMsg, 0 );
    return FALSE;
  }

  sprintf( cMsg, APPNAME " running on COM%d", iVcomPort );
  listMessage( cMsg, 0 );
  return TRUE;
}

/*----------------------------------------------------------------------
  processTimer(...)

  Processes the 1 second timer message

  returns: TRUE = OK, FALSE = Error
----------------------------------------------------------------------*/
static BOOL processTimer( HWND hwnd )
{
  static BOOL   bFirst=TRUE;
  BOOL          bRc;
  SYSTEMTIME    sysTime;
  char          cMsg[256];
  unsigned char ucBuf[48];
  unsigned      hh;
  int           dd, nn, mm;
  DWORD         dwRxBytes, dwBytesReceived, dwBytesWritten;
  FT_STATUS     ftStat;

  if ( !ftHandle )
  {
    // oops, something went badly wrong
    bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
    listMessage( "Too many errors, cancelling USB", 0 );
    return FALSE;
  }

  if ( iDateTimeFormat == DATETIMELOCAL )	// 20131218jb
  {
    GetLocalTime( &sysTime );
    hh = ( sysTime.wHour > 12 ) ? ( sysTime.wHour - 12 ) : sysTime.wHour;
    if ( !hh ) hh = 12;	// don't show 0 hours
    sprintf( cMsg, "%u : %02u : %02u %s", hh,
             (unsigned) sysTime.wMinute, (unsigned) sysTime.wSecond,
             ( sysTime.wHour > 11 ) ? "PM" : "AM" );
  }
  else
  {
	GetSystemTime( &sysTime );
    sprintf( cMsg, "%u : %02u : %02u UTC", (unsigned) sysTime.wHour,
             (unsigned) sysTime.wMinute, (unsigned) sysTime.wSecond );
  }

  // Build the time message for the remote
  nn = strlen( cMsg );
  ucBuf[0] = 'T';         // dislplay time on 1st line command
  ucBuf[1] = (unsigned char) ( nn + 3 );
  memcpy( ucBuf+2, cMsg, nn );
  nn += 2;
  ucBuf[nn] = checksum( ucBuf, nn );
  ++nn;

  ftStat = FT_Write( ftHandle, ucBuf, nn, &dwBytesWritten );
  if ( ftStat != FT_OK )
  {
    bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
    sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
    listMessage( cMsg, 0 );
    return FALSE;
  }

  if ( !sysTime.wSecond || bFirst )
  {
    // Sent only one a minute
    mm = ( sysTime.wMonth > 12 ) ? 0 : sysTime.wMonth;
	dd = ( sysTime.wDayOfWeek > 6 ) ? 7 : sysTime.wDayOfWeek;
    sprintf( cMsg, "%s %02u, %02u %s", pcMonth[mm],
             (unsigned) sysTime.wDay, (unsigned) sysTime.wYear, pcDay[dd] );

    // Build the date message for the micro
    nn = strlen( cMsg );
    ucBuf[0] = 'D';         // dislplay date on 2nd line command
    ucBuf[1] = (unsigned char) ( nn + 3 );
    memcpy( ucBuf+2, cMsg, nn );
    nn += 2;
    ucBuf[nn] = checksum( ucBuf, nn );
    ++nn;

    ftStat = FT_Write( ftHandle, ucBuf, nn, &dwBytesWritten );
    if ( ftStat != FT_OK )
    {
      bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
      sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
      listMessage( cMsg, 0 );
      return FALSE;
    }

    bFirst = FALSE;
  }

  if ( !sysTime.wMinute && !sysTime.wSecond )
  {
    // Build the beeper message for the micro
    ucBuf[0] = 'S';               // sound the beeper
    ucBuf[1] = 3;
    ucBuf[2] = checksum( ucBuf, 2 );

    ftStat = FT_Write( ftHandle, ucBuf, 3, &dwBytesWritten );
    if ( ftStat != FT_OK )
    {
      bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
      sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
      listMessage( cMsg, 0 );
      return FALSE;
    }
  }

  // Now see if any request has been received

  ftStat = FT_GetQueueStatus( ftHandle, &dwRxBytes );
  if ( ftStat != FT_OK )
  {
    bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
    sprintf( cMsg, "FT_GetQueueStatus failed, rc = %u", (unsigned) ftStat );
    listMessage( cMsg, 0 );
    return FALSE;
  }

  if ( dwRxBytes < 3 ) return TRUE;
  if ( dwRxBytes > sizeof( ucBuf ) )
    dwRxBytes = sizeof( ucBuf );

  ftStat = FT_Read( ftHandle, ucBuf, dwRxBytes, &dwBytesReceived );
  if ( ftStat != FT_OK )
  {
    bRc = KillTimer( hwnd, CLOCK_TIMER_ID );
    sprintf( cMsg, "FT_Read failed, rc = %u", (unsigned) ftStat );
    listMessage( cMsg, 0 );
    return FALSE;
  }

  if ( ucBuf[0] == 'Z' )        // shutdown request
  {
    if ( ucBuf[1] == 3 )
      if ( ucBuf[2] = ucBuf[0] + ucBuf[1] )
      {
        EndDialog( hwnd, 0 );
        return TRUE;
      }
  }
  else
    if ( ucBuf[0] == 'N' )      // id information
    {
      if ( dwBytesReceived == (DWORD) ucBuf[1] )
      {
        nn = dwBytesReceived - 3;
        memcpy( cMsg, ucBuf+2, nn );
        cMsg[nn] = '\0';
        listMessage( cMsg, 0 );
      }
    }
	else
      if ( ucBuf[0] == 'L' )      // local time
      {
        if ( dwBytesReceived == (DWORD) ucBuf[1] )
        {
		  iDateTimeFormat = DATETIMELOCAL;
		  bFirst = TRUE;          // force date to be sent too
		  listMessage( "Received request for local time", 0 );
        }
      }
	  else
        if ( ucBuf[0] == 'U' )      // UTC time
        {
          if ( dwBytesReceived == (DWORD) ucBuf[1] )
          {
		    iDateTimeFormat = DATETIMEUTC;
		    bFirst = TRUE;          // force date to be sent too
		    listMessage( "Received request for UTC time", 0 );
          }
        }

  return TRUE;
}

/*----------------------------------------------------------------------
  On_MYWM_NOTIFYICON(...)

  On_MYWM_NOTIFYICON - processes callback messages for taskbar icons.
  wParam - first message parameter of the callback message.
  lParam - second message parameter of the callback message.
----------------------------------------------------------------------*/
static void On_MYWM_NOTIFYICON( HWND hwnd, WPARAM wParam, LPARAM lParam )
{
  BOOL  bRc;
  UINT  uMouseMsg = (UINT) lParam;

  switch ( uMouseMsg )
  {
    case WM_LBUTTONDOWN:
    case WM_RBUTTONDOWN:
      bRc = ShowWindow( hwnd, /* SW_SHOW */  SW_RESTORE );
      bRc = SetForegroundWindow( hwnd );
  }
}

/*----------------------------------------------------------------------
  MyTaskBarIcon(...)

  Adds an icon to the taskbar status area.
  Returns TRUE if successful, or FALSE otherwise.
  hwnd - handle to the window to receive callback messages.
----------------------------------------------------------------------*/
static BOOL MyTaskBarIcon( HWND hwnd, HICON hIcon )
{
  BOOL           bRcode;
  NOTIFYICONDATA tnid;
  DWORD          dwMsg;

  tnid.cbSize = sizeof(NOTIFYICONDATA);
  tnid.hWnd = hwnd;
  tnid.uID = IDI_ICON;
  tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
  tnid.uCallbackMessage = UM_NOTIFYICON;
  tnid.hIcon = hIcon;
  strcpy( tnid.szTip, APPNAME );

  dwMsg = ( bIconAdded ) ? NIM_MODIFY : NIM_ADD;
  bRcode = Shell_NotifyIcon( dwMsg, &tnid);

  if ( bRcode )
    bIconAdded = TRUE;

  return bRcode;
}

/*----------------------------------------------------------------------
  MyTaskBarDeleteIcon(...)

  Deletes an icon from the taskbar status area.
  Returns TRUE if successful, or FALSE otherwise.
  hwnd - handle to the window that added the icon.
----------------------------------------------------------------------*/
static BOOL MyTaskBarDeleteIcon( HWND hwnd )
{
  BOOL bRcode;
  NOTIFYICONDATA tnid;

  tnid.cbSize = sizeof(NOTIFYICONDATA);
  tnid.hWnd = hwnd;
  tnid.uID = IDI_ICON;

  bRcode = Shell_NotifyIcon(NIM_DELETE, &tnid);
  return bRcode;
}

/*-------------------------------------------------------------------
  Add a message to the listbox
-------------------------------------------------------------------*/
void listMessage( char *pcMsg, int iOption )
{
  LRESULT            lRc;

  if ( !hwndDlg ) return;  // could use message box

  lRc = SendDlgItemMessage( hwndDlg, IDC_MSGS, LB_ADDSTRING, 0,
                            (LPARAM) (LPCTSTR) pcMsg );
}

/*-------------------------------------------------------------------
  Search for and open the requested FTDI COM port.

  Input: 1. bReport: FALSE = do not issue MessageBox for recoverable errors
                      TRUE = issue MessageBox for recoverable errors
         2. Global variable iVcomPort

  Output: If successful, sets global variable ftHandle

  Returns: -1 = unrecoverable error
            0 = recoverable error
			1 = successful
-------------------------------------------------------------------*/
static int getComPortHandle( BOOL bReport )
{
  FT_STATUS     ftStat;
  DWORD         dwArg1, dwArg2;
  int           ii;
  LONG          lComPortNumber;
  char          cMsg[92];

  // Get the number of USB devices available
  ftStat = FT_ListDevices( &dwArg1, &dwArg2, FT_LIST_NUMBER_ONLY );
  if ( ftStat != FT_OK )
  {
    sprintf( cMsg, "FT_ListDevices failed, rc = %d", (unsigned) ftStat );
    MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    return -1;
  }
  if ( !dwArg1 )     // number of devices found
  {
    if ( bReport )
      MessageBox( NULL, "FT_ListDevices reports no devices found.",
                  APPNAME, MB_OK | MB_ICONERROR );
    return 0;
  }

  // Search for the com port the user specified
  for ( ii=0; ii<(int) dwArg1; ++ii )
  {
    ftStat = FT_Open( ii, &ftHandle );
    if ( ftStat == FT_OK )
    {
      ftStat = FT_GetComPortNumber( ftHandle, &lComPortNumber );
      if ( ftStat != FT_OK )
      {
        ftStat = FT_Close( ftHandle );
        continue;
      }
      if ( lComPortNumber == iVcomPort )
        break;

      ftStat = FT_Close( ftHandle );
    }
  }

  if ( ii == (int) dwArg1 )
  {
    if ( bReport )
    {
      sprintf( cMsg, "Could not find COM%d in any of %d FTDI devices.",
               iVcomPort, dwArg1 );
      MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
    }
    return 0;
  }

  return 1;	    // success
}

/*-------------------------------------------------------------------
  Synchronize with the remote application program

  Sends the synchronize character and looks for correct response.
  Send x'ee', receive 'a'.

  Returns: True = OK, FALSE = ERROR
-------------------------------------------------------------------*/
static BOOL synchronize_avr( FT_HANDLE ftHandle )
{
  DWORD         dwBytesWritten, dwRxBytes, dwBytesReceived;
  FT_STATUS     ftStat;
  int           ii;
  unsigned char ucBuf[32];
  char          cMsg[128];

  for ( ii=0; ii<5; ++ii )
  {
    ucBuf[0] = 0xee;
    ftStat = FT_Write( ftHandle, ucBuf, 1, &dwBytesWritten );
    if ( ftStat != FT_OK )
    {
      sprintf( cMsg, "FT_Write failed, rc = %u", (unsigned) ftStat );
      MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
      return FALSE;
    }

    Sleep( 1000 );

    ftStat = FT_GetQueueStatus( ftHandle, &dwRxBytes );
    if ( ftStat != FT_OK )
    {
      sprintf( cMsg, "FT_GetQueueStatus failed, rc = %u", (unsigned) ftStat );
      MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );
      return FALSE;
    }

    if ( !dwRxBytes )
      continue;

    if ( dwRxBytes > sizeof( ucBuf ) )
      dwRxBytes = sizeof( ucBuf );

    ftStat = FT_Read( ftHandle, ucBuf, dwRxBytes, &dwBytesReceived );
    if ( ftStat != FT_OK )
    {
      sprintf( cMsg, "FT_Read failed, rc = %u", (unsigned) ftStat );
      MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );  
      return FALSE;
    }

    if ( ucBuf[0] == 'a' )
      return TRUE;

    ftStat = FT_Purge( ftHandle, ( FT_PURGE_RX | FT_PURGE_TX ) );
    if ( ftStat != FT_OK )
    {
      sprintf( cMsg, "FT_Purge failed, rc = %u", (unsigned) ftStat );
      MessageBox( NULL, cMsg, APPNAME, MB_OK | MB_ICONERROR );  
      return FALSE;
    }
  }

  return FALSE;
}

/*-------------------------------------------------------------------
  Compute the checksum

  Returns: checksum
-------------------------------------------------------------------*/
static unsigned char checksum( unsigned char *pucBuffer, int nBytes )
{
  int			ii;
  unsigned char ucCk=0;

  for ( ii=0; ii<nBytes; ++ii )
    ucCk += pucBuffer[ii];

  return ( ucCk & 0x7f );
}